#include <math.h>
#include "synth.h"



Sine::Sine(double _freq, double _attack , double _decay, double _sustain)
{
    freq = _freq;
    attack = _attack;
    decay = _decay;
    sustain = _sustain;
    if (!refcount)
        sintab = new SinTab;
    
    refcount++;

}

Sine::Sine(double rfreq, Sine const *other)
{
    freq = rfreq * other->freq;
    attack = other->attack;
    decay = other->decay;
    sustain = other->sustain;
    if (!refcount)
        sintab = new SinTab;
    
    refcount++;
}



Sine::~Sine()
{
    refcount--;

    if (!refcount)
        delete sintab;
}

int Sine::refcount = 0;
SinTab *Sine::sintab = NULL;


int Sine::render(int pos, int destfreq, int destbits)
{
    double t = ((double)pos) / destfreq;
    double v;

    if (t < attack)
        v =  t/attack;
    else if (t < attack+decay)
        v = 1 + (sustain -1) * ( t-attack) / decay;
    else
        v = sustain;
    
    v *= (destbits == 8) ? 127 : 32767;


    return (int)(v * (sintab->sin(t * freq)));
}

SinTab::SinTab()
{
    for (int i=0; i< SINTABSIZE; i++)
        sintab[i] = ::sin((double)i * 2 * M_PI/ SINTABSIZE);
}

// sine, one period in f (0,1]
double SinTab::sin(double f)
{
    f = f - (int)f;
    f *= SINTABSIZE;


    return sintab[(int)f];
}

Synth::Synth(int max_sines)
{
    max_nr_sines = max_sines;
    nr_sines = 0;

    sines = new Sine *[max_sines];
    volumes = new double[max_sines];


    state = STATE_STOPPED;
    stream = NULL;

}

Synth::Synth(double relfreq , Synth const * other)
{
    int i;
    max_nr_sines = other->max_nr_sines;
    nr_sines = other->nr_sines;

    sines = new Sine *[max_nr_sines];
    volumes = new double[max_nr_sines];

    for (i = 0; i< nr_sines; i++)
    {
        volumes[i] = other->volumes[i];
        sines[i] = new Sine(relfreq, other->sines[i]);
        
    }

    state = STATE_STOPPED;
    stream = play_audio_stream(SYNTH_BUF, 16,0, SYNTH_FREQ,255, 128);
}



Synth::~Synth()
{
    int i;
    for (i = 0 ; i < nr_sines; i++)
        delete sines[i];

    delete [] sines;
    delete [] volumes;
}

void Synth::add_sine(double freq, double attack, double decay, double sustain, double vol)
{
    ASSERT(nr_sines < max_nr_sines);

    sines[nr_sines] = new Sine(freq, attack, decay, sustain);
    volumes[nr_sines] = vol;
    nr_sines++;
    
}

void Synth::start(double r, double v,int _vibrato_time, int _vibrato_depth)
{

    voice_ramp_volume(stream->voice, 100, 255);

    volume = v;
    relfreq = r;
    count = 0;
    update();
    state = STATE_PLAYING;

    lastvibtime = 0;
    vibdir = 0;

    vibtime = _vibrato_time;
    vibdepth = _vibrato_depth;
    
}

int Synth::update()
{
    int i, j;
    double val;
    if (!stream)
        return 0;

    int t = count * 1000 / SYNTH_FREQ;

    if (vibtime)
    {
        if (t - lastvibtime > vibtime)
        {
            lastvibtime = t;
            vibdir = !vibdir;
            voice_sweep_frequency(stream->voice, vibtime, SYNTH_FREQ + (vibdir ?  vibdepth : - vibdepth));
            
        }

    }
    

    unsigned short *buffer = (unsigned short *)get_audio_stream_buffer(stream);

    if (!buffer)
        return 1;


    for (i = 0; i < SYNTH_BUF; i++)
    {
        val = 0;
        count++;

        if (state != STATE_STOPPED)
        {
            for (j = 0; j< nr_sines;j++)
                val += sines[j]->render((int)(count * relfreq), SYNTH_FREQ, 16) * volumes[j];

            val /= max_nr_sines;
            val *= volume;

        }
        else
            val = 0;

        val += 32768;

        ASSERT(val >= 0);
        ASSERT(val < 65536);


        buffer[i] = (unsigned short)val;
    }
    
    free_audio_stream_buffer(stream);

    if (state == STATE_STOPPING)
    {
        buffers_to_go--;
        if (!buffers_to_go)
        {
            state = STATE_STOPPED;

            return 0;
        }
    }
    
    return 1;

}

void Synth::stop()
{

        
    ASSERT(stream);

    voice_ramp_volume(stream->voice, STOPPING_RAMP, 0);
    buffers_to_go = STOPPING_BUFFERS;
    state = STATE_STOPPING;

}


int Synth::get_synthfreq()
{
    return SYNTH_FREQ;
}
